@@ -29,6 +29,7 @@ import pickle
 import shlex
 import shutil
 import signal
+import socket
 import subprocess
 import tempfile
 from collections import defaultdict
@@ -147,13 +148,6 @@ class Qtile(CommandObject):
         for button in self.config.mouse:
             self.grab_button(button)
 
-        # no_spawn is set after the very first startup; we only want to run the
-        # startup hook once.
-        if not self.no_spawn:
-            hook.fire("startup_once")
-            self.no_spawn = True
-        hook.fire("startup")
-
         if self._state:
             if isinstance(self._state, str):
                 try:
@@ -169,6 +163,18 @@ class Qtile(CommandObject):
 
         self.core.on_config_load(initial)
 
+        if self.config.reconfigure_screens:
+            hook.subscribe.screen_change(self.reconfigure_screens)
+
+        # no_spawn is set after the very first startup; we only want to run the
+        # startup hook once. Note that this needs to happen *after* the config
+        # is loaded and we have subscribed reconfigure_screens() in case people
+        # do xrandr style manipulation in this hook.
+        if not self.no_spawn:
+            hook.fire("startup_once")
+            self.no_spawn = True
+        hook.fire("startup")
+
         if self._state:
             for screen in self.screens:
                 screen.group.layout.show(screen.get_rect())
@@ -177,9 +183,6 @@ class Qtile(CommandObject):
         self.update_desktops()
         hook.subscribe.setgroup(self.update_desktops)
 
-        if self.config.reconfigure_screens:
-            hook.subscribe.screen_change(self.reconfigure_screens)
-
         # Start the sleep inhibitor process to listen to sleep signals
         # NB: the inhibitor will only connect to the dbus service if the
         # user has used the "suspend" or "resume" hooks in their config.
@@ -220,15 +223,20 @@ class Qtile(CommandObject):
         faulthandler.register(signal.SIGUSR2, all_threads=True)
 
         try:
+            signals: dict[signal.Signals, Callable]
+            signals = {
+                signal.SIGTERM: self.stop,
+                signal.SIGINT: self.stop,
+                signal.SIGHUP: self.stop,
+                signal.SIGUSR1: self.reload_config,
+            }
+            if self.core.name == "x11":
+                # the wayland backend installs its own SIGCHLD handler after
+                # the XWayland X server has initialized (as a workaround). the
+                # x11 backend can just do it here.
+                signals[signal.SIGCHLD] = utils.reap_zombies
             async with (
-                LoopContext(
-                    {
-                        signal.SIGTERM: self.stop,
-                        signal.SIGINT: self.stop,
-                        signal.SIGHUP: self.stop,
-                        signal.SIGUSR1: self.reload_config,
-                    }
-                ),
+                LoopContext(signals),
                 ipc.Server(
                     self._prepare_socket_path(self.socket_path),
                     self.server.call,
@@ -325,8 +333,7 @@ class Qtile(CommandObject):
                     layout.finalize()
 
             for screen in self.screens:
-                for gap in screen.gaps:
-                    gap.finalize()
+                screen.finalize_gaps()
         except:  # noqa: E722
             logger.exception("exception during finalize")
         hook.clear()
@@ -731,28 +738,24 @@ class Qtile(CommandObject):
         if self.current_screen and isinstance(win, base.Window):
             # Window may have been bound to a group in the hook.
             if not win.group and self.current_screen.group:
-                self.current_screen.group.add(win, focus=win.can_steal_focus)
+                self.current_screen.group.add(win)
 
         hook.fire("client_managed", win)
 
     def unmanage(self, wid: int) -> None:
         c = self.windows_map.get(wid)
         if c:
-            group = None
+            # Fire the hook before removing the group from the window so hooked
+            # functions can access it
+            hook.fire("client_killed", c)
             if isinstance(c, base.Static):
                 if c.reserved_space:
                     self.free_reserved_space(c.reserved_space, c.screen)
             elif isinstance(c, base.Window):
                 if c.group:
-                    group = c.group
                     c.group.remove(c)
             del self.windows_map[wid]
 
-            if isinstance(c, base.Window):
-                # Put the group back on the window so hooked functions can access it.
-                c.group = group
-            hook.fire("client_killed", c)
-
     def find_screen(self, x: int, y: int) -> Screen | None:
         """Find a screen based on the x and y offset"""
         result = []
@@ -819,6 +822,12 @@ class Qtile(CommandObject):
                 closest_screen = s
         return closest_screen or self.screens[0]
 
+    def _focus_hovered_window(self) -> None:
+        window = self.core.hovered_window
+        if window:
+            if isinstance(window, base.Window):
+                window.focus()
+
     def process_button_click(self, button_code: int, modmask: int, x: int, y: int) -> bool:
         handled = False
         for m in self._mouse_map[button_code]:
@@ -826,6 +835,8 @@ class Qtile(CommandObject):
                 continue
 
             if isinstance(m, Click):
+                if self.config.follow_mouse_focus == "click_or_drag_only":
+                    self._focus_hovered_window()
                 for i in m.commands:
                     if i.check(self):
                         status, val = self.server.call(
@@ -837,6 +848,8 @@ class Qtile(CommandObject):
             elif (
                 isinstance(m, Drag) and self.current_window and not self.current_window.fullscreen
             ):
+                if self.config.follow_mouse_focus == "click_or_drag_only":
+                    self._focus_hovered_window()
                 if m.start:
                     i = m.start
                     status, val = self.server.call((i.selectors, i.name, i.args, i.kwargs, False))
@@ -1263,7 +1276,11 @@ class Qtile(CommandObject):
 
     @expose_command()
     def spawn(
-        self, cmd: list[str] | str, shell: bool = False, env: dict[str, str] = dict()
+        self,
+        cmd: list[str] | str,
+        shell: bool = False,
+        env: dict[str, str] = dict(),
+        group: str | None = None,
     ) -> int:
         """
         Spawn a new process.
@@ -1277,6 +1294,8 @@ class Qtile(CommandObject):
             -c". This enables the use of shell syntax within the command (e.g. pipes).
         env:
             Dictionary of environmental variables to pass with command.
+        group:
+            Name of group in which to spawn process.
 
         Examples
         ========
@@ -1301,82 +1320,72 @@ class Qtile(CommandObject):
             logger.error("couldn't find `%s`", to_lookup)
             return -1
 
-        r, w = os.pipe()
-        pid = os.fork()
-        if pid < 0:
-            os.close(r)
-            os.close(w)
-            return pid
+        if len(env) == 0:
+            env = os.environ.copy()
+            # if qtile was installed in a virutal env, we don't
+            # necessarily want to propagate that to children
+            # applications, since it may change e.g. the behavior
+            # of shells that spawn python applications
+            env.pop("VIRTUAL_ENV", None)
 
-        if pid == 0:
-            os.close(r)
+        # std{in,out,err} should be /dev/null
+        with open("/dev/null") as null:
+            file_actions: list[tuple] = [
+                (os.POSIX_SPAWN_DUP2, 0, null.fileno()),
+                (os.POSIX_SPAWN_DUP2, 1, null.fileno()),
+                (os.POSIX_SPAWN_DUP2, 2, null.fileno()),
+            ]
 
-            # close qtile's stdin, stdout, stderr so the called process doesn't
-            # pollute our xsession-errors.
-            os.close(0)
-            os.close(1)
-            os.close(2)
+            # To create a rule to move process to a specific group and manage race conditions
+            # we need to do the following:
+            # 1) Create a socket pair so we can communicate to child process
+            # 2) Spawn a "qtile launch" process and pass the file descriptor
+            # 3) Have "qtile launch" notify qtile that it's ready
+            # 4) Create a rule using the pid from step 2
+            # 5) Send a message over the socket to "qtile launch"
+            # 6) Once the message is received, "qtile launch" execs the desired process
+            if group is not None:
+                parent_sock, child_sock = socket.socketpair()
+                # socket pair fds are non-inheritable by default but we need this to survice
+                os.set_inheritable(child_sock.fileno(), True)
 
-            pid2 = os.fork()
-            if pid2 == 0:
-                os.close(w)
-                try:
-                    # if qtile was installed in a virutal env, we don't
-                    # necessarily want to propagate that to children
-                    # applications, since it may change e.g. the behavior
-                    # of shells that spawn python applications
-                    del os.environ["VIRTUAL_ENV"]
-                except KeyError:
-                    pass
+            # this API is only available on python 3.13 or above, and only on platforms
+            # where posix_spawn_file_actions_addclosefrom_np() exists (only glibc on Linux).
+            # where we're creating a rule to manage the group, we'll handle closing fds in
+            # `qtile launch`
+            if hasattr(os, "POSIX_SPAWN_CLOSEFROM") and group is None:
+                # we should close all fds so that child processes don't
+                # accidentally write to our x11 event loop or whatever; we never
+                # used to do this, so it seems fine to only do it where this nice
+                # API to do it exists.
+                file_actions.append((os.POSIX_SPAWN_CLOSEFROM, 3))
 
-                for k, v in env.items():
-                    os.environ[k] = v
+            try:
+                if group is not None:
+                    # As per step 2 (above) we want to launch the process via "qtile launch"
+                    args = ["qtile", "launch", "--fd", f"{child_sock.fileno()}"] + args
+                pid = os.posix_spawnp(args[0], args, env, file_actions=file_actions)
+                if group is not None:
+                    # Listen for READY message from `qtile launch`
+                    data = parent_sock.recv(5)
 
-                # Open /dev/null as stdin, stdout, stderr
-                try:
-                    fd = os.open(os.devnull, os.O_RDWR)
-                except OSError:
-                    # This shouldn't happen, catch it just in case
-                    pass
-                else:
-                    # For Python >=3.4, need to set file descriptor to inheritable
-                    try:
-                        os.set_inheritable(fd, True)
-                    except AttributeError:
-                        pass
+                    if data != b"READY":
+                        logger.warning("Received unexpected data: {data!r}")
 
-                    # Again, this shouldn't happen, but we should just check
-                    if fd > 0:
-                        os.dup2(fd, 0)
+                    # Create the group matching rule
+                    match_args = {"net_wm_pid": pid}
+                    rule_args = {"group": group, "one_time": True}
+                    self.add_rule(match_args=match_args, rule_args=rule_args)
 
-                    os.dup2(fd, 1)
-                    os.dup2(fd, 2)
+                    # Tell child process it can now spawn main proces as rule has been added
+                    parent_sock.send(b"OK")
 
-                try:
-                    os.execvp(args[0], args)
-                except OSError:
-                    # can't log here since we forked :(
-                    pass
-
-                os._exit(1)
-            else:
-                # Here it doesn't matter if fork failed or not, we just write
-                # its return code and exit.
-                os.write(w, str(pid2).encode())
-                os.close(w)
-
-                # sys.exit raises SystemExit, which will then be caught by our
-                # top level catchall and we'll end up with two qtiles; os._exit
-                # actually calls exit.
-                os._exit(0)
-        else:
-            os.close(w)
-            os.waitpid(pid, 0)
-
-            # 1024 bytes should be enough for any pid. :)
-            pid = int(os.read(r, 1024))
-            os.close(r)
-            return pid
+                    parent_sock.close()
+                    child_sock.close()
+                return pid
+            except OSError as e:
+                logger.warning("failed to execute: %s: %s", str(args), str(e))
+                return -1
 
     @expose_command()
     def status(self) -> Literal["OK"]:
